# pkg.nu -- 包管理器的抽象
#
# Copyright (C) 2022 KAAAsS
use pkgs/pacman.nu
use pkgs/homebrew.nu
use pkgs/apt.nu

export-env {
    # 所有可用的包管理器
    $env.package_managers = [
        (pacman pack),
        (homebrew pack),
        (apt pack),
        (fallback_manager)
    ]
    # 计算默认包管理器
    $env.default_package_manager = (select_managers)
}

def select_managers [
    manager?: string,
    --allow-unavailable
] {
    if ($manager == null) {
        # 选择第一个可用的包管理器
        for $it in $env.package_managers {
            let checker = ($it | get "available?")

            if (do $checker) {
                return $it
            }
        }

        error make {
            msg: "没有可用的包管理器"
        }
    } else {
        # 选择对应名称的包管理器
        let ret = ($env.package_managers | filter {|it|
            ($it.name == $manager)
        })

        # 未找到
        if ($ret | is-empty) {
            error make {
                msg: $"未找到名称为 ($manager) 的包管理器"
            }
        }

        # 检查是否可用
        let ret = ($ret | get 0)
        let checker = ($ret | get "available?")

        if (not (do $checker)) {
            if ($allow_unavailable) {
                return null
            } else {
                error make {
                    msg: $"包管理器 ($manager) 不可用"
                }
            }
        }

        return $ret
    }
}

# Fallback 包管理器，无法实际承担包管理功能，仅用于占位
def fallback_manager [] {
    {
        "name": "fallback",
        "available?": {
            log warn "无可用的包管理器！若安装包则会发生错误！"
            true
        },
        "check_install?": {|p| false },
        "install": {|p| false },
        "uninstall": {|p| false }
    }
}

# 检查包是否安装
export def check_install? [
    pkg_name: string,
    --manager: string
] {
    let manager_impl = (select_managers $manager)
    let func = ($manager_impl | get "check_install?")
    do $func $pkg_name
}

# 安装包
export def install [
    pkg_name: string,
    --manager: string
] {
    use log.nu

    if (check_install? --manager=$manager $pkg_name) {
        log info $"包 ($pkg_name) 已经安装，跳过"
    } else {
        log info $"安装包 ($pkg_name)..."
        # 安装包
        let manager_impl = (select_managers $manager)
        let func = ($manager_impl | get "install")
        let ret = (do $func $pkg_name)
        # 检查安装结果
        if ($ret) {
            log info $"成功安装包 ($pkg_name)"
        } else {
            error make {
                msg: $"无法安装包 ($pkg_name)",
                label: {
                    text: "包名",
                    span: (metadata $pkg_name).span
                }
            }
        }
    }
}

# 卸载包
export def uninstall [
    pkg_name: string,
    --manager: string
] {
    use log.nu

    if (not (check_install? --manager=$manager $pkg_name)) {
        log info $"未安装包 ($pkg_name)，跳过"
    } else {
        log info $"卸载包 ($pkg_name)..."
        # 卸载包
        let manager_impl = (select_managers $manager)
        let func = ($manager_impl | get "uninstall")
        let ret = (do $func $pkg_name)
        # 检查安装结果
        if ($ret) {
            log info $"成功卸载包 ($pkg_name)"
        } else {
            error make {
                msg: $"无法卸载包 ($pkg_name)",
                label: {
                    text: "包名",
                    span: (metadata $pkg_name).span
                }
            }
        }
    }
}

# 解析包对象对应的包管理器
export def resolve_manager [
    pkg_desc
] {
    let type = ($pkg_desc | describe)

    if ($type == "string") {
        # 未指定特殊包管理器，直接使用默认
        return {
            "pkg": $pkg_desc,
            "manager": $env.default_package_manager.name
        }
    } else {
        # 存在特殊包管理器配置，就遍历 key 直到找到可用的包管理器
        for $expect_manager in ($pkg_desc | transpose key).key {
            let manager_impl = (select_managers --allow-unavailable $expect_manager)
            
            if ($manager_impl != null) {
                return {
                    "pkg": ($pkg_desc | get $expect_manager),
                    "manager": $manager_impl.name
                }
            }
        }

        return null
    }
}

export def test [] {

    # 增加 Mock 管理器
    $env.package_managers = [
        {
            "name": "unavailable",
            "available?": { false }
        },
        {
            "name": "mock",
            "available?": { true },
            "check_install?": {|p| $p == "mock" or $p == "mockmock" },
            "install": {|p| $p == "mock" },
            "uninstall": {|p| $p == "mock" }
        },
        {
            "name": "mock2",
            "available?": { true },
            "check_install?": {|p| $p == "mock2" },
        }
    ]
    $env.default_package_manager = (select_managers)

    # select_managers
    let mock = (select_managers "mock2")
    assert ($mock.name == "mock2")
    let mock = (select_managers)
    assert ($mock.name == "mock")
    assert error {
        select_managers "unavailable"
    }

    # check_install?
    assert (check_install? "mock")
    assert (not (check_install? "mock2"))
    assert (check_install? --manager=mock2 "mock2")
    assert (not (check_install? --manager=mock2 "mock"))

    # install
    install "mock"
    assert error {
        install "mock2"
    }

    # uninstall
    uninstall "mock"
    uninstall "mock2" # 未安装
    assert error {
        uninstall "mockmock"
    }

    # resolve_manager
    assert ((resolve_manager "test").manager == "mock")
    assert ((resolve_manager {
        "mock": "test2",
        "mock2": "test3"
    }).manager == "mock")
    assert ((resolve_manager {
        "mock2": "test3"
    }).manager == "mock2")
    assert error {
        resolve_manager {
            "mock3": "test3"
        }
    }
}
